from __future__ import annotations

import os
import gi  # type: ignore

from typing import Any, Optional, Dict, Callable

from .db import pwgen, DEFAULT_NEW_PASSWORD_OCTETS, Database

gi.require_version("Gtk", "4.0")
from gi.repository import Gtk  # type: ignore # noqa: E402
from gi.repository import GObject  # noqa: E402
from gi.repository import Gdk  # noqa: E402
from gi.repository import Gio  # noqa: E402
from gi.repository import GLib  # noqa: E402


############################################################


class ImpassContext(GObject.Object):  # type: ignore[misc]
    ctx = GObject.Property(type=str)

    def __init__(self, ctx: str, date: str) -> None:
        super().__init__()
        self.ctx = ctx
        self.date = date


class ImpassContextSorter(Gtk.Sorter):  # type: ignore[misc]
    def do_compare(self, a: ImpassContext, b: ImpassContext) -> Gtk.Ordering:
        if a.ctx < b.ctx:
            return Gtk.Ordering.SMALLER
        elif a.ctx == b.ctx:
            return Gtk.Ordering.EQUAL
        return Gtk.Ordering.LARGER


class ImpassListItemFactory(Gtk.SignalListItemFactory):  # type: ignore[misc]
    def __init__(self, gui: Gui) -> None:
        super().__init__()
        self.gui = gui
        self.connect("setup", ImpassListItemFactory.setup_hook)
        self.connect("bind", ImpassListItemFactory.bind_hook)

    def setup_hook(self, item: Gtk.ListItem) -> None:
        label = Gtk.Label(halign=Gtk.Align.START, hexpand=True)
        item.set_child(label)

    def bind_hook(self, item: Gtk.ListItem) -> None:
        label: Gtk.Label = item.get_child()
        ctx: ImpassContext = item.get_item()
        label.set_label(ctx.ctx)
        label.set_tooltip_text(ctx.date)


class Gui:
    """Impass GTK-based query UI."""

    def __init__(self, db: Database, query: Optional[str] = None) -> None:
        """
        +--------------------- warning --------------------+
        |                    notification                  |
        +--------------------------------------------------+
        |                    description                   |
        +----------------- simplebox ----------------------+
        | [_simplectxentry____] <simplebtn> <simplemenubtn>|
        +------------------- ctxbox -----------------------+
        |          | +----- ctxbox2 -----------------------+
        | ctxlabel | | [_ctxentry________________________] |
        |          | +---------------- --------------------+ (ctxwarninglabel shows when
        |          | | ctxwarninglabel                     |  ctxentry matches existing
        |          | +-------------------------------------+  entry (createbtn is also
        +------------------- passbox ----------------------+  disabled in this case))
        |           +------ passbox2 --------+             |
        | passlabel | [_passentry__________] | <createbtn> | createbtn saves, emits, and
        |           | passdescription        |             | closes
        +-----------+------------------------+-------------+
        """
        self.db = db
        self.query = query
        self.selected: Optional[Dict[str, str]] = None
        self.window: Gtk.Window
        self.entry: Gtk.Entry
        self.label: Gtk.Label
        self.app: Gtk.Application
        if query is not None:
            query = query.strip()

        if query:
            # If we have an intial query, directly do a search without
            # initializing any Gtk objects.  This will initialize the
            # database and potentially return entries.
            r = self.db.search(query)
            # If only a single entry is found, _search() will set the
            # result and attempt to close any X objects (of which
            # there are none).  Since we don't need to initialize any
            # GUI, return the initialization immediately.
            # See .returnValue().
            if len(r) == 1:
                self.selected = r[list(r.keys())[0]]
                return

        self.app = Gtk.Application(application_id="net.cmrg.impass")
        self.app.connect("activate", self.on_activate)

    def on_activate(self, app: Gtk.Application) -> None:
        self.window = Gtk.ApplicationWindow(application=app)
        self.window.set_title("🔒Impass Password Manager")

        self.frame = Gtk.Box(name="frame", orientation=Gtk.Orientation.VERTICAL)
        self.window.set_child(self.frame)

        self.css = Gtk.CssProvider()
        self.css.load_from_string(
            """
        label.warning {
          color: red;
        }
        label.info {
          font-style: italic;
        }
        popover.nondefaultaction contents {
          border-top-left-radius: 0px;
          border-top-right-radius: 0px;
        }
        popover.completion contents {
          border-radius: 0px;
          padding: 1px;
        }
        """
        )

        self.warning = Gtk.Box(
            name="warning",
            orientation=Gtk.Orientation.VERTICAL,
            visible=False,
        )
        self.frame.append(self.warning)
        self.notification = Gtk.Label(
            name="notification",
            css_classes=["warning"],
            label="⚠ could not validate signature on db file!",
        )
        self.warning.append(self.notification)
        self.warning.append(Gtk.Separator())

        self.description = Gtk.Label(
            name="description",
            label="Global state of impass gui",
        )
        self.frame.append(self.description)
        self.simplebox = Gtk.Box(name="simplebox")
        self.frame.append(self.simplebox)
        self.entry = Gtk.Entry(
            name="simplectxentry",
            hexpand=True,
            width_chars=50,
            placeholder_text="Enter context…",
        )
        self.simplebox.append(self.entry)
        self.simplebtn = Gtk.Button(name="simplebtn", label="Emit")
        self.simplebox.append(self.simplebtn)
        self.simplemenubtn = Gtk.MenuButton(name="simplemenubtn")
        self.simplebox.append(self.simplemenubtn)

        self.ctxbox = Gtk.Box(name="ctxbox", visible=False)
        self.frame.append(self.ctxbox)
        self.ctxlabel = Gtk.Label(name="ctxlabel", label="Context:")
        self.ctxbox.append(self.ctxlabel)
        self.ctxbox2 = Gtk.Box(name="ctxbox2", orientation=Gtk.Orientation.VERTICAL)
        self.ctxbox.append(self.ctxbox2)
        self.ctxentry = Gtk.Entry(
            name="ctxentry",
            hexpand=True,
            placeholder_text="Enter context…",
        )
        self.ctxbox2.append(self.ctxentry)
        self.ctxwarninglabel = Gtk.Label(
            name="ctxwarninglabel",
            visible=False,
            css_classes=["warning"],
            label="⚠ This context already exists!",
        )
        self.ctxbox2.append(self.ctxwarninglabel)

        self.passbox = Gtk.Box(name="passbox", visible=False)
        self.frame.append(self.passbox)
        self.passlabel = Gtk.Label(name="passlabel", label="Password:")
        self.passbox.append(self.passlabel)
        self.passbox2 = Gtk.Box(name="passbox2", orientation=Gtk.Orientation.VERTICAL)
        self.passbox.append(self.passbox2)
        self.passentry = Gtk.Entry(
            name="passentry",
            hexpand=True,
            visibility=False,
            primary_icon_name="view-refresh",
            secondary_icon_name="edit-find",
            primary_icon_tooltip_text="generate a new password",
            secondary_icon_tooltip_text="show_password",
            placeholder_text="You must enter a password",
            input_purpose=Gtk.InputPurpose.PASSWORD,
        )
        self.passbox2.append(self.passentry)
        self.passdescription = Gtk.Label(
            name="passdescription",
            css_classes=["info"],
            label="%d characters, %d lowercase, etc…",
        )
        self.passbox2.append(self.passdescription)
        self.createbtn = Gtk.Button(
            name="createbtn",
            label="Create and emit",
            receives_default=True,
        )
        self.passbox.append(self.createbtn)

        self.completion = Gtk.Popover(
            name="completion",
            position=Gtk.PositionType.BOTTOM,
            halign=Gtk.Align.START,
            has_arrow=False,
            css_classes=["completion", "background"],
            autohide=False,
        )
        self.simplebox.append(self.completion)

        delaction = self.add_action("delete", self.deleteclicked)
        customaction = self.add_action("createcustom", self.customclicked)
        refreshpassaction = self.add_action("refreshpass", self.refreshpass)
        showpassaction = self.add_action("showpass", self.showhidepass)
        simpleaction = self.add_action("simple", self.simpleclicked)
        createaction = self.add_action("create", self.create)

        emitmenu = Gio.Menu()
        emitmenu.insert(0, "Delete", "win.delete")
        self.emitmenu = Gtk.PopoverMenu(
            name="emitmenu",
            menu_model=emitmenu,
            has_arrow=False,
            halign=Gtk.Align.END,
            css_classes=["nondefaultaction", "background"],
            position=Gtk.PositionType.BOTTOM,
        )

        createmenu = Gio.Menu()
        createmenu.insert(0, "Create custom…", "win.createcustom")
        self.createmenu = Gtk.PopoverMenu(
            name="createmenu",
            menu_model=createmenu,
            has_arrow=False,
            halign=Gtk.Align.END,
            css_classes=["nondefaultaction", "background"],
            position=Gtk.PositionType.BOTTOM,
        )

        self.warning.set_visible(not self.db.sigvalid)

        self.passmenu = Gio.Menu()
        self.passmenu.insert(0, "_Generate a new password", "win.refreshpass")
        self.passmenu.insert(1, "_Show password", "win.showpass")
        self.passentry.set_extra_menu(self.passmenu)

        self.eck = Gtk.EventControllerKey()
        self.eck.connect("key-pressed", self.keypress)
        self.window.add_controller(self.eck)
        Gtk.StyleContext().add_provider_for_display(
            self.window.get_display(),
            self.css,
            Gtk.STYLE_PROVIDER_PRIORITY_USER,
        )

        self.window.connect("destroy", self.destroy)
        self.entry.connect("activate", self.simpleclicked)
        self.entry.connect("changed", self.update_simple_context_entry)
        self.simplebtn.connect("clicked", self.simpleclicked)
        self.ctxentry.connect("changed", self.update_ctxentry)
        self.ctxentry.connect("activate", self.customcreateclicked)
        self.passentry.connect("changed", self.update_passentry)
        self.passentry.connect("activate", self.customcreateclicked)
        self.passentry.connect("icon-press", self.passentry_icon_clicked)
        self.createbtn.connect("clicked", self.customcreateclicked)

        self.ctxlist = Gio.ListStore()
        for ctx in self.db:
            self.ctxlist.append(ImpassContext(ctx, self.db[ctx]["date"]))

        self.completionfilter = Gtk.CustomFilter()
        self.selector = Gtk.SingleSelection()
        self.completion_filter_model = Gtk.FilterListModel(
            model=self.ctxlist,
            filter=self.completionfilter,
        )
        self.sorted_model = Gtk.SortListModel(
            model=self.completion_filter_model,
            sorter=ImpassContextSorter(),
        )
        self.selector.set_model(self.sorted_model)
        self.completionlist = Gtk.ListView(
            model=self.selector,
            factory=ImpassListItemFactory(self),
            single_click_activate=True,
        )
        self.completion.set_child(self.completionlist)
        self.completionlist.connect(
            "activate",
            lambda listview, pos: self.completion_clicked(
                listview.get_model().get_item(pos).ctx
            ),
        )

        if self.query:
            self.entry.set_text(self.query)
        self.set_state("Enter context for desired password:")
        self.update_simple_context_entry(None)
        self.window.present()

    def add_action(
        self,
        label: str,
        func: Callable[[Optional[Gio.Action], Optional[GLib.Variant]], None],
    ) -> Gio.Action:
        ret = Gio.SimpleAction.new(label, None)
        ret.connect("activate", func)
        self.window.add_action(ret)
        return ret

    def set_state(self, state: str) -> None:
        self.description.set_label(state)

    def completion_clicked(self, label: str) -> None:
        self.entry.set_text(label)
        self.completion.set_visible(False)
        self.entry.grab_focus()

    def update_simple_context_entry(self, widget: Optional[Gtk.Widget]) -> None:
        sctx = self.entry.get_text().strip()
        max_matches = 40
        matches = []

        if sctx is not None and sctx != "":
            matches = sorted(filter(lambda x: sctx.lower() in x.lower(), self.db))[
                :max_matches
            ]
        if len(matches) == 0:
            self.completion.set_visible(False)
        else:
            lsctx = sctx.lower()
            self.completionfilter.set_filter_func(lambda x: lsctx in x.ctx.lower())
            self.completion.set_visible(True)

        menu = Gio.Menu()
        if sctx in self.db:
            self.simplebtn.set_label("Emit")
            self.simplemenubtn.set_popover(self.emitmenu)
            menu.insert(0, f"Emit password for '{sctx}'", "win.simple")
            menu.insert(1, f"Delete password for '{sctx}'", "win.delete")
        elif sctx is None or sctx == "":
            self.simplebtn.set_label("Create…")
            self.simplemenubtn.set_popover(None)
            menu.insert(0, "Create custom password…", "win.createcustom")
        else:
            self.simplebtn.set_label("Create")
            self.simplemenubtn.set_popover(self.createmenu)
            menu.insert(0, f"Create and emit password for '{sctx}'", "win.create")
            menu.insert(1, f"Create custom password for '{sctx}'", "win.createcustom")
        self.entry.set_extra_menu(menu)

    def simpleclicked(
        self, action: Optional[Gio.Action] = None, param: Optional[GLib.Variant] = None
    ) -> None:
        sctx = self.entry.get_text().strip()
        if sctx in self.db:
            self.selected = self.db[sctx]
            if self.selected is None:
                self.label.set_text(
                    "weird -- no context found even though there should be one"
                )
            else:
                self.window.close()
        elif sctx is None or sctx == "":
            self.customclicked(None)
        else:
            self.create(None)

    def keypress(
        self,
        keycontroller: Gtk.EventControllerKey,
        key: int,
        code: int,
        mods: Gdk.ModifierType,
    ) -> None:
        if key == Gdk.KEY_Escape:
            if self.completion.get_visible():
                self.completion.set_visible(False)
            else:
                self.app.quit()

    def create(
        self,
        action: Optional[Gio.Action] = None,
        param: Optional[GLib.Variant] = None,
    ) -> None:
        sctx = self.entry.get_text().strip()
        self.selected = self.db.add(sctx)
        self.db.save()
        self.window.close()

    def deleteclicked(
        self,
        action: Optional[Gio.Action] = None,
        param: Optional[GLib.Variant] = None,
    ) -> None:
        sctx = self.entry.get_text().strip()
        confirmation = Gtk.MessageDialog(
            transient_for=self.window,
            modal=True,
            buttons=Gtk.ButtonsType.OK_CANCEL,
            message_type=Gtk.MessageType.QUESTION,
            text="Are you sure you want to delete the password for '" + sctx + "'?",
        )

        def delok(_: Optional[Gtk.Widget]) -> None:
            self.db.remove(sctx)
            self.db.save()
            confirmation.close()
            self.window.close()

        ok = confirmation.get_widget_for_response(Gtk.ResponseType.OK)
        ok.connect("clicked", delok)
        cancel = confirmation.get_widget_for_response(Gtk.ResponseType.CANCEL)
        cancel.connect("clicked", lambda x: confirmation.close())
        confirmation.present()

    def customclicked(
        self,
        action: Optional[Gio.Action] = None,
        param: Optional[GLib.Variant] = None,
    ) -> None:
        if self.ctxentry is None or self.entry is None:
            raise Exception("Gui is not initialized")
        self.simplebox.set_visible(False)
        self.ctxbox.set_visible(True)
        self.passbox.set_visible(True)
        self.ctxentry.set_text(self.entry.get_text())
        selection = self.entry.get_selection_bounds()
        if selection:
            self.ctxentry.select_region(selection[0], selection[1])
        else:
            self.ctxentry.set_position(self.entry.get_position())
        self.ctxentry.grab_focus_without_selecting()
        self.set_state("Create new password (with custom settings):")
        self.refreshpass()
        self.update_ctxentry()

    def update_ctxentry(
        self, widget: Optional[Gtk.Widget] = None, data: Optional[Any] = None
    ) -> None:
        sctx = self.ctxentry.get_text().strip()
        if sctx in self.db:
            self.ctxwarninglabel.set_visible(True)
            self.ctxwarninglabel.set_text("The context '%s' already exists!" % (sctx))
            self.createbtn.set_sensitive(False)
        elif sctx is None or sctx == "":
            self.ctxwarninglabel.set_visible(False)
            self.createbtn.set_sensitive(False)
        else:
            self.ctxwarninglabel.set_visible(False)
            self.createbtn.set_sensitive(self.passentry.get_text() != "")

    def update_passentry(
        self, widget: Optional[Gtk.Widget] = None, data: Optional[Any] = None
    ) -> None:
        newpass = self.passentry.get_text()
        sctx = self.ctxentry.get_text().strip()
        ln = len(newpass)
        # FIXME: should check (and warn) for non-ascii characters
        lcount = len("".join(filter(lambda x: x.islower(), newpass)))
        ucount = len("".join(filter(lambda x: x.isupper(), newpass)))
        ncount = len("".join(filter(lambda x: x.isnumeric(), newpass)))
        ocount = ln - (lcount + ucount + ncount)
        desc = "%d characters (%d lowercase, %d uppercase, %d number, %d other)" % (
            ln,
            lcount,
            ucount,
            ncount,
            ocount,
        )
        self.createbtn.set_sensitive(
            newpass != "" and sctx != "" and sctx not in self.db
        )
        self.passdescription.set_text(desc)

    def passentry_icon_clicked(
        self,
        widget: Gtk.Widget,
        pos: Gtk.EntryIconPosition,
        data: Optional[Any] = None,
    ) -> None:
        if pos == Gtk.EntryIconPosition.PRIMARY:
            self.refreshpass()
        elif pos == Gtk.EntryIconPosition.SECONDARY:
            self.showhidepass()

    def showhidepass(
        self,
        action: Optional[Gio.Action] = None,
        param: Optional[GLib.Variant] = None,
    ) -> None:
        newvis = not self.passentry.get_visibility()
        self.passentry.set_visibility(newvis)
        label = "Hide password" if newvis else "Show password"
        self.passentry.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, label)
        self.passmenu.remove(1)
        self.passmenu.insert(1, f"_{label}", "win.showpass")

    def refreshpass(
        self,
        action: Optional[Gio.Action] = None,
        param: Optional[GLib.Variant] = None,
    ) -> None:
        pwsize = os.environ.get("IMPASS_PASSWORD", DEFAULT_NEW_PASSWORD_OCTETS)
        try:
            pwsize = int(pwsize)
        except ValueError:
            pwsize = DEFAULT_NEW_PASSWORD_OCTETS
        newpw = pwgen(pwsize)
        self.passentry.set_text(newpw)
        # FIXME: should refocus self.passentry?

    def customcreateclicked(
        self, widget: Optional[Gtk.Widget] = None, event: Optional[Gdk.Event] = None
    ) -> None:
        newctx = self.ctxentry.get_text().strip()
        newpass = self.passentry.get_text()
        if newpass == "" or newctx == "" or newctx in self.db:
            # this button is not supposed to work under these conditions
            return
        self.selected = self.db.add(newctx, password=newpass)
        self.db.save()
        self.window.close()

    def destroy(self, widget: Gtk.Widget, data: Optional[Any] = None) -> None:
        self.window.close()

    def return_value(self) -> Optional[Dict[str, str]]:
        if self.selected is None:
            self.app.run()
        return self.selected



import os
import io
import stat
import json
import gpg  # type: ignore
import codecs
import datetime

from typing import Optional, Dict, Iterator

############################################################

DEFAULT_NEW_PASSWORD_OCTETS = 18


def pwgen(nbytes: int) -> str:
    """Return *nbytes* bytes of random data, base64-encoded."""
    s = os.urandom(nbytes)
    b = codecs.encode(s, "base64")
    b = bytes(filter(lambda x: x not in b"=\n", b))
    return codecs.decode(b, "ascii")


############################################################


class DatabaseError(Exception):
    def __init__(self, msg: str) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return repr(self.msg)


class Database:
    """An impass database."""

    def __init__(
        self, dbpath: Optional[str] = None, keyid: Optional[str] = None
    ) -> None:
        """Database at dbpath will be decrypted and loaded into memory.

        If dbpath is not specified, an empty database will be
        initialized.

        The sigvalid property is set False if any OpenPGP signatures
        on the db file are invalid.  sigvalid is None for new
        databases.

        """
        self._dbpath = dbpath
        self._keyid = keyid

        # default database information
        self._type = "impass"
        self._version = 1
        self._entries: Dict[str, Dict[str, str]] = {}

        self._gpg = gpg.Context()
        self._gpg.armor = True
        self._sigvalid: Optional[bool] = None

        if self._dbpath and os.path.exists(self._dbpath):
            try:
                cleardata = self._decrypt_db(self._dbpath)
                # FIXME: trap exception if json corrupt
                jsondata = json.loads(cleardata.decode("utf-8"))
            except IOError as e:
                raise DatabaseError(str(e))

            # unpack the json data
            # FIXME: we accept "assword" type for backwords compatibility
            if "type" not in jsondata or jsondata["type"] not in [
                self._type,
                "assword",
            ]:
                raise DatabaseError("Database is not a proper impass database.")
            if "version" not in jsondata or jsondata["version"] != self._version:
                raise DatabaseError("Incompatible database.")
            self._entries = jsondata["entries"]

    @property
    def version(self) -> int:
        """Database version."""
        return self._version

    @property
    def sigvalid(self) -> Optional[bool]:
        """Validity of OpenPGP signature on db file."""
        return self._sigvalid

    def __str__(self) -> str:
        return '<impass.Database "%s">' % (self._dbpath)

    def __repr__(self) -> str:
        return 'impass.Database("%s")' % (self._dbpath)

    def __getitem__(self, context: str) -> Dict[str, str]:
        """Return database entry for exact context."""
        return self._entries[context]

    def __contains__(self, context: str) -> bool:
        """True if context string in database."""
        return context in self._entries

    def __iter__(self) -> Iterator[str]:
        """Iterator of all database contexts."""
        return iter(self._entries)

    def _decrypt_db(self, path: str) -> bytes:
        data = None
        self._sigvalid = False
        with open(path, "rb") as f:
            try:
                data, _, vfy = self._gpg.decrypt(f, verify=True)
                for s in vfy.signatures:
                    if s.validity >= gpg.constants.VALIDITY_FULL:
                        self._sigvalid = True
            except gpg.errors.GPGMEError:
                # retry decryption without verification:
                with open(path, "rb") as try2:
                    data, _, _ = self._gpg.decrypt(try2, verify=False)
        if not isinstance(data, bytes):
            raise DatabaseError(
                f"expected gpg.Context.decrypt() to return bytes, got {type(data)}"
            )
        return data

    def _encrypt_db(self, data: io.BytesIO, keyid: Optional[str]) -> bytes:
        # The signer and the recipient are assumed to be the same.
        # FIXME: should these be separated?
        try:
            recipient = self._gpg.get_key(keyid or self._keyid, secret=False)
            signer = self._gpg.get_key(keyid or self._keyid, secret=False)
        except gpg.errors.GPGMEError:
            raise DatabaseError("Could not retrieve GPG encryption key.")
        self._gpg.signers = [signer]
        data.seek(0)
        encdata, _, _ = self._gpg.encrypt(
            data, [recipient], always_trust=True, compress=False
        )
        if not isinstance(encdata, bytes):
            raise DatabaseError(
                f"expected gpg.Context.decrypt() to return bytes, got {type(data)}"
            )
        return encdata

    def _set_entry(
        self, context: str, password: Optional[str] = None
    ) -> Dict[str, str]:
        if not isinstance(password, str):
            if password is None:
                bytes = DEFAULT_NEW_PASSWORD_OCTETS
            if isinstance(password, int):
                bytes = password
            password = pwgen(bytes)
        e = {"password": password, "date": datetime.datetime.utcnow().isoformat() + "Z"}
        self._entries[context] = e
        return e

    def add(self, context: str, password: Optional[str] = None) -> Dict[str, str]:
        """Add new entry.

        If password is None, one will be generated automatically.  If
        password is an int it will be interpreted as the number of
        random bytes to use.

        If the context is already in the db a DatabaseError will be
        raised.

        Database changes are not saved to disk until the save() method
        is called.

        """
        if context == "":
            raise DatabaseError("Can not add empty string context")
        if context in self:
            raise DatabaseError("Context already exists (see replace())")
        return self._set_entry(context, password)

    def replace(self, context: str, password: Optional[str] = None) -> Dict[str, str]:
        """Replace entry password.

        If password is None, one will be generated automatically.  If
        password is an int it will be interpreted as the number of
        random bytes to use.

        If the context is not in the db a DatabaseError will be
        raised.

        Database changes are not saved to disk until the save() method
        is called.

        """
        if context not in self:
            raise DatabaseError("Context not found (see add())")
        return self._set_entry(context, password)

    def update(self, old_context: str, new_context: str) -> None:
        """Update entry context.

        If the old context is not in the db a DatabaseError will be
        raised.

        Database changes are not saved to disk until the save() method
        is called.

        """
        if old_context not in self:
            raise DatabaseError("Context '%s' not found." % old_context)
        password = self[old_context]["password"]
        self.add(new_context, password)
        self.remove(old_context)

    def remove(self, context: str) -> None:
        """Remove entry.

        Database changes are not saved to disk until the save() method
        is called.

        """
        if context not in self:
            raise DatabaseError("Context '%s' not found" % context)
        del self._entries[context]

    def save(self, keyid: Optional[str] = None, path: Optional[str] = None) -> None:
        """Save database to disk.

        Key ID must either be specified here or at database initialization.
        If path not specified, database will be saved at original dbpath location.

        """
        # FIXME: should check that recipient is not different than who
        # the db was originally encrypted for
        if not keyid:
            keyid = self._keyid
        if not keyid:
            raise DatabaseError("Key ID for decryption not specified.")
        if not path:
            path = self._dbpath
        if not path:
            raise DatabaseError("Save path not specified.")
        jsondata = {
            "type": self._type,
            "version": self._version,
            "entries": self._entries,
        }
        cleardata = io.BytesIO(json.dumps(jsondata, indent=2).encode("utf-8"))
        encdata = self._encrypt_db(cleardata, keyid)
        newpath = path + ".new"
        bakpath = path + ".bak"
        mode = stat.S_IRUSR | stat.S_IWUSR
        with open(newpath, "wb") as f:
            f.write(encdata)
        if os.path.exists(path):
            mode = os.stat(path)[stat.ST_MODE]
            os.rename(path, bakpath)
        # FIXME: shouldn't this be done when we're actually creating
        # the file?
        os.chmod(newpath, mode)
        os.rename(newpath, path)

    def search(self, string: Optional[str] = None) -> Dict[str, Dict[str, str]]:
        """Search for string in contexts.

        If query is None, all entries will be returned.

        """
        mset = {}
        for context, entry in self._entries.items():
            # simple substring match
            if not string or string in context:
                mset[context] = entry
        return mset



#!/usr/bin/env python3

# much of the structure here was cribbed from
# https://github.com/pypa/sampleproject

from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

version = {}
with open("impass/version.py") as f:
    exec(f.read(), version)

setup(
    name="impass",
    version=version["__version__"],
    description="Simple and secure password management system.",
    long_description=long_description,
    author="Jameson Rollins",
    author_email="jrollins@finestructure.net",
    url="https://salsa.debian.org/debian/impass",
    license="GPLv3+",
    packages=["impass"],
    keywords=["passwords password-management"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        # maybe extend this?  "impass gui" won't work on anything but
        # X11, but the rest of it might still be useful.
        "Environment :: X11 Applications",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3 :: Only",
    ],
    install_requires=[
        "gpg",
        "PyGobject",
    ],
    # You can install this optional dependency using the following syntax:
    # $ pip install -e .xdo
    extras_require={
        "xdo": ["xdo"],
    },
    # https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/
    # should we have a 'gui_scripts' as well?
    entry_points={
        "console_scripts": [
            "impass = impass.__main__:main",
        ],
    },
)



#!/usr/bin/env python3

import os
import io
import sys
import json
import gpg  # type: ignore
import getpass
import argparse
import textwrap
import subprocess
import collections

from typing import Optional, NoReturn, List, Callable, Union, Sequence, Any, Dict

from .db import Database, DatabaseError, DEFAULT_NEW_PASSWORD_OCTETS
from .version import __version__

PROG = "impass"
_swaymark = f"🔐{PROG}"

############################################################

IMPASS_DIR = os.path.join(os.path.expanduser("~"), ".impass")

############################################################


def xclip(text: str) -> None:
    p = subprocess.Popen(" ".join(["xclip", "-i"]), shell=True, stdin=subprocess.PIPE)
    p.communicate(text.encode("utf-8"))
    # FIXME: shouldn't we warn the user if xclip fails here?


def log(*args: str) -> None:
    print(*args, file=sys.stderr)


############################################################
# Return codes:
# 1 command/load line error
# 2 context/password invalid
# 5 db doesn't exist
# 10 db error
# 20 gpg/key error
############################################################
def error(code: int, msg: str = "") -> NoReturn:
    if msg:
        log(msg)
    sys.exit(code)


def open_db(keyid: Optional[str] = None, create: bool = False) -> Database:
    db_path = os.getenv("IMPASS_DB", os.path.join(IMPASS_DIR, "db"))
    if not create and not os.path.exists(db_path):
        error(
            5,
            """Impass database does not exist.
To add an entry to the database use 'impass add'.
See 'impass help' for more information.""",
        )
    try:
        db = Database(db_path, keyid)
    except gpg.errors.GPGMEError as e:
        error(20, "Decryption error: {}".format(e))
    except DatabaseError as e:
        error(10, "Impass database error: {}".format(e.msg))
    if db.sigvalid is False:
        log("WARNING: could not validate OpenPGP signature on db file.")
    return db


def get_keyid() -> str:
    keyid = os.getenv("IMPASS_KEYID")
    keyfile = os.getenv("IMPASS_KEYFILE", os.path.join(IMPASS_DIR, "keyid"))

    if not keyid and os.path.exists(keyfile):
        with open(keyfile, "r") as f:
            keyid = f.read().strip()

    save = False
    if keyid is None or keyid == "":
        log("OpenPGP key ID of encryption target not specified.")
        log("Please provide key ID in IMPASS_KEYID environment variable,")
        log("or specify key ID now to save in ~/.impass/keyid file.")
        keyid = input("OpenPGP key ID: ")
        if keyid == "":
            keyid = None
        else:
            save = True

    if keyid is None or keyid == "":
        error(20)

    try:
        gpgctx = gpg.Context()
        gpgctx.get_key(keyid, secret=False)
    except gpg.errors.GPGMEError as e:
        log("GPGME error for key ID {}:".format(keyid))
        log("  {}".format(e))
        error(20)

    if save:
        if not os.path.isdir(os.path.dirname(keyfile)):
            os.mkdir(os.path.dirname(keyfile))
        with open(keyfile, "w") as f:
            f.write(keyid)

    return keyid


class Completer:
    def __init__(self, completions: Optional[List[str]] = None):
        if completions is None:
            self.completions: List[str] = []
        else:
            self.completions = completions

    def completer(self, text: str, index: int) -> Optional[str]:
        matching = [c for c in self.completions if c.startswith(text)]
        try:
            return matching[index]
        except IndexError:
            return None


def input_complete(
    prompt: str, completions: Optional[List[str]] = None, default: Optional[str] = None
) -> str:
    try:
        # lifted from magic-wormhole/codes.py
        import readline

        c = Completer(completions)
        readline.set_startup_hook()
        if "libedit" in readline.__doc__:
            readline.parse_and_bind("bind ^I rl_complete")
        else:
            readline.parse_and_bind("tab: complete")
        readline.set_completer(c.completer)
        readline.set_completer_delims(" ")
        if default:
            readline.set_startup_hook(lambda: readline.insert_text(default))
    except ImportError:
        pass
    try:
        return input(prompt)
    except KeyboardInterrupt:
        error(-1)


def retrieve_context(
    arg: Optional[str],
    prompt: str = "context: ",
    default: Optional[str] = None,
    stdin: bool = True,
    db: Optional[Database] = None,
) -> str:
    if arg == "-" and stdin:
        context = sys.stdin.read()
    elif arg is None or arg == ":":
        if db:
            context = input_complete(
                prompt, completions=[c for c in db], default=default
            )
        else:
            context = input_complete(prompt, default=default)
    else:
        context = arg
    return context.strip()


class PasswordAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Optional[Union[str, Sequence[Any]]],
        option_string: Optional[str] = None,
    ) -> None:
        env_password = os.getenv("IMPASS_PASSWORD")
        if env_password is None or env_password == "":
            password: Optional[Union[str, int]] = None
        elif env_password in ["prompt", ":"]:
            password = ":"
        else:
            try:
                password = int()
            except ValueError:
                error(
                    1,
                    "IMPASS_PASSWORD environment variable is neither int nor 'prompt'.",
                )
        if values == ":":
            password = ":"
        elif isinstance(values, str):
            try:
                password = int(values)
            except ValueError:
                error(666, "Don't type your password on the command line!!!")
        setattr(namespace, self.dest, password)


def retrieve_password(pwspec: str) -> str:
    if pwspec == ":":
        return input_password()
    else:
        log("Auto-generating password...")
        return pwspec


def input_password() -> str:
    try:
        password0 = getpass.getpass("password: ")
        password1 = getpass.getpass("reenter password: ")
        if password0 != password1:
            error(2, "Passwords do not match.  Aborting.")
        return password0
    except KeyboardInterrupt:
        error(-1)


############################################################
# command functions


def add(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Add new entry.

    If the context already exists in the database an error will be
    thrown.

    """
    parser = argparse.ArgumentParser(
        prog=PROG + " add",
        description=add.__doc__,
    )
    parser.add_argument(
        "context",
        nargs="?",
        help="existing database context, ':' for prompt, or '-' for stdin",
    )
    parser.add_argument(
        "pwspec",
        nargs="?",
        action=PasswordAction,
        help="password spec: N octets or ':' for prompt",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)

    keyid = get_keyid()
    db = open_db(keyid, create=True)

    context = retrieve_context(argsns.context)
    if context in db:
        error(2, "Context '{}' already exists.".format(context))

    password = retrieve_password(argsns.pwspec)

    try:
        db.add(context, password)
        db.save()
    except DatabaseError as e:
        error(10, "Impass database error: {}".format(e.msg))
    log("New entry writen.")
    return parser


def replace(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Replace password for entry.

    If the context does not already exist in the database an error
    will be thrown.

    """
    parser = argparse.ArgumentParser(
        prog=PROG + " replace",
        description=replace.__doc__,
    )
    parser.add_argument(
        "context",
        nargs="?",
        help="existing database context, ':' for prompt, or '-' for stdin",
    )
    parser.add_argument(
        "pwspec",
        nargs="?",
        action=PasswordAction,
        help="password spec: N octets or ':' for prompt",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)

    keyid = get_keyid()
    db = open_db(keyid)

    context = retrieve_context(argsns.context, db=db)
    if context not in db:
        error(2, "Context '{}' not found.".format(context))

    password = retrieve_password(argsns.pwspec)

    try:
        db.replace(context, password)
        db.save()
    except DatabaseError as e:
        error(10, "Impass database error: {}".format(e.msg))
    log("Password replaced.")
    return parser


def update(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Update context for existing entry, keeping password the same.

    Special context value of '-' can only be provided to the old
    context.

    """
    parser = argparse.ArgumentParser(
        prog=PROG + " update",
        description=update.__doc__,
    )
    parser.add_argument(
        "old_context",
        nargs="?",
        help="existing database context, ':' for prompt, or '-' for stdin",
    )
    parser.add_argument(
        "new_context",
        nargs="?",
        help="new database context or ':' for prompt",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)

    keyid = get_keyid()
    db = open_db(keyid)

    old_context = retrieve_context(argsns.old_context, prompt="old context: ", db=db)
    if old_context not in db:
        error(2, "Context '{}' not found".format(old_context))

    new_context = retrieve_context(
        argsns.new_context, prompt="new context: ", default=old_context, stdin=False
    )
    if new_context in db:
        error(2, "Context '{}' already exists.".format(new_context))

    try:
        db.update(old_context, new_context)
        db.save()
    except DatabaseError as e:
        error(10, "Impass database error: {}".format(e.msg))
    log("Entry updated.")
    return parser


def dump(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Dump password database to stdout as json.

    If a string is provide only entries whose context contains the
    string will be dumped. Otherwise all entries are returned.
    Passwords will not be displayed unless IMPASS_DUMP_PASSWORDS is
    set.

    """
    parser = argparse.ArgumentParser(prog=PROG + " dump", description=dump.__doc__)
    parser.add_argument(
        "string",
        nargs="?",
        help="substring match for contexts",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)
    keyid = get_keyid()
    db = open_db(keyid)
    results = db.search(argsns.string)
    output: Dict[str, Dict[str, str]] = {}
    for context in results:
        output[context] = {}
        output[context]["date"] = results[context]["date"]
        if os.getenv("IMPASS_DUMP_PASSWORDS"):
            output[context]["password"] = results[context]["password"]
    print(json.dumps(output, sort_keys=True, indent=2))
    return parser


def get(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Print password matching context to stdout."""
    parser = argparse.ArgumentParser(
        prog=PROG + " get",
        description=get.__doc__,
    )
    parser.add_argument(
        "context",
        nargs="?",
        help="exact context match, ':' for prompt, or '-' for stdin",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)
    keyid = get_keyid()
    db = open_db(keyid)
    context = retrieve_context(argsns.context, db=db)
    if context not in db:
        error(2, "Context '{}' not found".format(context))
    print(db[context]["password"])
    return parser


def gui(
    args: Optional[List[str]],
    method: Optional[str] = os.getenv("IMPASS_XPASTE", None),
) -> argparse.ArgumentParser:
    """Launch minimal GUI.

    Good for X11 or Wayland-based window manager integration.
    Upon invocation the user
    will be prompted to decrypt the database, after which a graphical
    search prompt will be presented. If an additional string is
    provided, it will be added as the initial search string. All
    matching results for the query will be presented to the user.
    When a result is selected, the password will be retrieved
    according to the method specified by IMPASS_XPASTE. If no match
    is found, the user has the opportunity to generate and store a new
    password, which is then delivered via IMPASS_XPASTE.

    Note: contexts that have leading or trailing whitespace are not
    accessible through the GUI.

    """
    parser = argparse.ArgumentParser(
        prog=PROG + " gui",
        description=gui.__doc__,
    )
    parser.add_argument(
        "string",
        nargs="?",
        help="substring match for contexts",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)
    from .gui import Gui

    if method is None:
        if os.getenv("SWAYSOCK", None) is not None:
            method = "sway"
        elif os.getenv("DISPLAY", None) is not None:
            method = "xdo"

    if method == "xdo":
        try:
            import xdo  # type: ignore
        except ModuleNotFoundError:
            error(
                1,
                """The xdo module is not found, so 'xdo' pasting is not available.
                Please install python3-xdo.""",
            )
        # initialize xdo
        x = xdo.xdo()
        # get the id of the currently focused window
        win = x.get_focused_window()
    elif method == "xclip":
        pass
    elif method == "sway":
        try:
            import i3ipc  # type: ignore
        except ModuleNotFoundError:
            error(
                1,
                """The i3ipc module is not found, so 'sway' pasting is not available.
                Please install python3-i3ipc.""",
            )
        i3conn = i3ipc.Connection()
        res = i3conn.command(f"mark {_swaymark}")
        if len(res) != 1 and not res[0]["success"]:
            error(1, "Failed to mark focused window")
        con_info = i3conn.get_tree().find_focused()
        if _swaymark not in con_info.marks:
            error(1, "The focused window was not marked")
        criteria = f"con_mark={_swaymark} con_id={con_info.id} pid={con_info.pid}"
    else:
        error(1, "Unknown X paste method '{}'.".format(method))
    keyid = get_keyid()
    db = open_db(keyid)
    result = Gui(db, query=argsns.string).return_value()
    # type the password in the saved window
    if result:
        if method == "xdo":
            x.focus_window(win)
            x.wait_for_window_focus(win)
            x.type(result["password"])
        elif method == "xclip":
            xclip(result["password"])
        elif method == "sway":
            # pick the right element
            i3conn.command(f"[{criteria}] focus")
            proc = subprocess.Popen(
                ["wtype", "-"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            (stdout, stderr) = proc.communicate(result["password"].encode())
            i3conn.command(f"[{criteria}] unmark")
            if proc.returncode:
                error(1, "failed to run wtype to inject keystrokes")
        else:
            error(1, f"Unknown X paste method '{method}'.")
    else:
        if method == "sway":
            i3conn.command(f"[{criteria}] unmark")

    return parser


def remove(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Remove entry.

    If the context does not already exist in the database an error
    will be thrown.

    """
    parser = argparse.ArgumentParser(
        prog=PROG + " remove",
        description=remove.__doc__,
    )
    parser.add_argument(
        "context",
        nargs="?",
        help="existing database context, ':' for prompt, or '-' for stdin",
    )
    if args is None:
        return parser
    argsns = parser.parse_args(args)

    keyid = get_keyid()
    db = open_db(keyid)

    context = retrieve_context(argsns.context, db=db)
    if context not in db:
        error(2, "Context '{}' not found.".format(context))

    try:
        log("Really remove entry '{}'?".format(context))
        response = input("Type 'yes' to remove: ")
    except KeyboardInterrupt:
        error(-1)
    if response != "yes":
        error(-1)

    try:
        db.remove(context)
        db.save()
    except DatabaseError as e:
        error(10, "Impass database error: {}".format(e.msg))
    log("Entry removed.")
    return parser


def print_help(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Full usage or command help (also '-h' after command)."""
    parser = argparse.ArgumentParser(
        prog=PROG + " help",
        description=print_help.__doc__,
    )
    if args is None:
        return parser
    # if no argument is provided print the full man page
    try:
        cmd = args[0]
    except IndexError:
        print_manpage()
        return parser
    # otherwise assume the first argument is a command and print it's
    # help
    func = get_func(cmd)
    func(["-h"])
    return parser


def version(args: Optional[List[str]]) -> argparse.ArgumentParser:
    """Print version."""
    parser = argparse.ArgumentParser(
        prog=PROG + " version",
        description=version.__doc__,
    )
    if args is None:
        return parser
    print(__version__)
    return parser


############################################################
# main

synopsis = f"{PROG} <command> [<args>...]"


# NOTE: double spaces are interpreted by text2man to be paragraph
# breaks.  NO DOUBLE SPACES.  Also two spaces at the end of a line
# indicate an element in a tag list.
def print_manpage() -> None:
    print(
        f"""
NAME
  {PROG} - Simple and secure password management and retrieval system

SYNOPSIS
  {synopsis}

DESCRIPTION

  The password database is stored as a single json object, OpenPGP
  encrypted and signed, and written to local disk (see
  IMPASS_DB). The file is created upon addition of the first
  entry. Database entries are keyed by 'context'. During retrieval of
  passwords the database is decrypted and read into memory. Contexts
  are searched by sub-string match.

  Contexts can be any string. If a context string is not specified on
  the command line it can be provided at a prompt, which features tab
  completion for contexts already in the database. One may also
  specify a context of '-' to read the context from stdin, or ':' to
  force a prompt.

  Passwords are auto-generated by default with {DEFAULT_NEW_PASSWORD_OCTETS} bytes of
  entropy. The number of octets can be specified with the
  IMPASS_PASSWORD environment variable or via the 'pwspec' optional
  argument to relevant commands. The length of the actually generated
  password will sometimes be longer than the specified bytes due to
  base64 encoding. If pwspec is ':' the user will be prompted for the
  password.

COMMANDS

{format_commands(man=True)}

SIGNATURES
    During decryption, OpenPGP signatures on the db file are checked
    for validity. If any of them are found to not be valid, a warning
    message will be written to stderr.

ENVIRONMENT
    IMPASS_DB  
        Path to impass database file. Default: ~/.impass/db

    IMPASS_KEYFILE  
        File containing OpenPGP key ID of database encryption
        recipient. Default: ~/.impass/keyid

    IMPASS_KEYID  
        OpenPGP key ID of database encryption recipient. This
        overrides IMPASS_KEYFILE if set.

    IMPASS_PASSWORD  
        See Passwords above.

    IMPASS_DUMP_PASSWORDS  
        Include passwords in dump when set.

    IMPASS_XPASTE  
        Method for password retrieval from GUI. Options are: 'xdo',
        which attempts to type the password into the window that had
        focus on launch, 'xclip' which inserts the password in the X
        clipboard, and 'sway', which types the password into the
        focused wayland container. Default: xdo or sway, detected
        automatically.

AUTHOR
    Jameson Graef Rollins <jrollins@finestructure.net>
    Daniel Kahn Gillmor <dkg@fifthhorseman.net>
""".strip()  # noqa: W291 (trailing whitespace)
    )


def format_commands(man: bool = False) -> str:
    prefix = " " * 8
    wrapper = textwrap.TextWrapper(
        width=70,
        initial_indent=prefix,
        subsequent_indent=prefix,
    )
    with io.StringIO("some initial text data") as f:
        for name, func in CMDS.items():
            if man:
                parser = func(None)
                if parser is None:
                    raise Exception(
                        f"{name} yielded function {func} that did"
                        "not return a parser (internal error)"
                    )
                startlen = len("usage: impass ")
                usage = parser.format_usage()[startlen:].strip()
                if parser.description is None:
                    desc = ""
                else:
                    desc = wrapper.fill(
                        "\n".join(
                            [ln.strip() for ln in parser.description.splitlines() if ln]
                        )
                    )
                f.write("  {}  \n".format(usage))
                f.write(desc + "\n")
                f.write("\n")
            else:
                if func.__doc__ is not None:
                    desc = func.__doc__.splitlines()[0]
                    f.write("  {:15}{}\n".format(name, desc))
        output = f.getvalue()
    return output.rstrip()


CMDS: collections.OrderedDict[
    str, Callable[[Optional[List[str]]], argparse.ArgumentParser]
] = collections.OrderedDict(
    [
        ("add", add),
        ("replace", replace),
        ("update", update),
        ("dump", dump),
        ("get", get),
        ("gui", gui),
        ("remove", remove),
        ("help", print_help),
        ("version", version),
    ]
)

ALIAS = {
    "--version": "version",
    "--help": "help",
    "-h": "help",
}


def get_func(cmd: str) -> Callable[[Optional[List[str]]], argparse.ArgumentParser]:
    """Retrieve the appropriate function from the command argument."""
    if cmd in ALIAS:
        cmd = ALIAS[cmd]
    try:
        return CMDS[cmd]
    except KeyError:
        log("Unknown command: {}".format(cmd))
        log("See 'help' for usage.")
        error(1)


def main() -> None:
    if len(sys.argv) < 2:
        log("Command not specified.")
        log("usage: {}".format(synopsis))
        log()
        log(format_commands())
        error(1)

    cmd = sys.argv[1]
    args = sys.argv[2:]
    func = get_func(cmd)

    # DEPRECATE: this is for the assword->impass transition
    if os.path.basename(sys.argv[0]) == "assword":
        log("WARNING: assword has been renamed impass. Please update your invocations.")
    vfound = []
    for var in ["DB", "KEYFILE", "KEYID", "PASSWORD", "DUMP_PASSWORDS", "XPASTE"]:
        val = os.getenv("ASSWORD_" + var)
        if val:
            vfound.append(var)
            if not os.getenv("IMPASS_" + var):
                os.environ["IMPASS_" + var] = val
    if vfound:
        log(
            """WARNING: assword has been renamed impass.
        Please update your environment variables:"""
        )
        for var in vfound:
            log("  ASSWORD_{var} -> IMPASS_{var}".format(var=var))
    assword_dir = os.path.join(os.path.expanduser("~"), ".assword")
    if (
        os.path.exists(assword_dir)
        and (not os.path.islink(assword_dir))
        and os.path.isdir(assword_dir)
        and (not os.getenv("IMPASS_DB"))
        and cmd not in ["help", "version", "-h", "--help", "--version"]
    ):
        try:
            os.rename(assword_dir, IMPASS_DIR)
            linkerr: Optional[str] = None
            try:
                os.symlink(IMPASS_DIR, assword_dir)
            except Exception as e:
                linkerr = str(e)
                pass
            print("renamed ~/.assword -> ~/.impass", file=sys.stderr)
            if linkerr is not None:
                linkerr = f"(failed symlinking ~/.assword to ~/.impass ({linkerr}))"
                print(linkerr, file=sys.stderr)
        except Exception as e:
            sys.exit(
                f"""Could not rename old assword directory ~/.assword -> ~/.impass.
            Error: {e}
            Please check ~/.impass path."""
            )
        os.symlink(IMPASS_DIR, assword_dir)
        log("renamed ~/.assword -> ~/.impass")
    # DEPRECATE

    cmd = sys.argv[1]
    args = sys.argv[2:]
    func = get_func(cmd)
    func(args)


if __name__ == "__main__":
    main()
